from functools import partial
from munch import Munch

from infi.shell.commands import commandgroup, command
from infi.shell.types import YesNo
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments, CommandAborted, CommandFailed

from ...utils import yesno
from ..remote import get_remote_system
from ..exceptions import ManagementException
from .hierarchical_base import HierarchicalEntityCommandsBase
from .replication import ReplicationCommandsMixin
from .types import (
    Str,
    Name,
    Pool,
    ConsistencyOrSnapGroup,
    ConsistencyGroup,
    SnapGroup,
    MasterVolume,
    VolumeSnap,
    Dataset,
    ReplicationLink,
    Replica,
    ReplicaRoleOrNone,
    )
from .utils import get_target_pool
from .options import QueryOptions
from . import render_collection, roles, LazyIdNameMapping, set_renderer_query_fields

QUERY_SUBSTITUTES = dict(
    members_count="members",
    is_replicated="replicated",
    )

DATASET_QUERY_FIELDS =          strlist("name size used allocated mapped")
DATASET_DETAILED_QUERY_FIELDS = strlist("name serial dataset_type thin size used allocated tree_allocated pool write_protected mapped ssd_enabled replication_role created_at updated_at")

DATASET_DETAILED_QUERY_SUBSTITUTES = dict(
    ssd_enabled="ssd_cache",
    write_protected="write_protected"
    )


class RemoteEntity(Str):
    def get_description(self):
        return "Name of a remote entity (auto-completion unavailable)"

class ConsistencyGroupCommandsBase(HierarchicalEntityCommandsBase):
    HIERARCHICAL_TYPE = ConsistencyOrSnapGroup
    _MEMBER_TYPES = dict(VOLUME="volume", FILESYSTEM="filesystem")
    _MEMBER_HIERARCHY_TYPES = dict(MASTER="", SNAP=" snapshot", CLONE=" clone")

    def _query_hook(self, cgs, pool_names_by_id=None):
        if pool_names_by_id:
            for cg in cgs:
                cg.pool = pool_names_by_id[cg.pool_id]
        return cgs

    def _get_member_type_name(self, member):
        return "{}{}".format(
            self._MEMBER_TYPES.get(member.dataset_type, "<unknown volume type>"),
            self._MEMBER_HIERARCHY_TYPES.get(member.type, ""))

def create_common_cg_commands_mixin(CgType, MemberType):

    class CommonCgCommandsMixin(object):
        @command(
            description="Delete {}".format(CgType.DESCRIPTION_SUBJECT),
            types=(CgType.List, ),
            substitute=dict(
                cgs="name",
                ),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def delete(self, _api, cgs):
            delete_members=False
            return self._delete(_api, cgs, params=dict(delete_members=delete_members))

        @command(
            description="List all cg members",
            types=(CgType.List,),
            substitute=dict(
                cg="name",
                ),
            options=QueryOptions(fields=DATASET_DETAILED_QUERY_FIELDS, sortable=DATASET_DETAILED_QUERY_FIELDS, substitute=DATASET_DETAILED_QUERY_SUBSTITUTES, detailed=True),
            )
        @render_collection(
            entity=Dataset,
            fields=DATASET_QUERY_FIELDS,
            detailed_fields=DATASET_DETAILED_QUERY_FIELDS,
            detailed_substitute=DATASET_DETAILED_QUERY_SUBSTITUTES,
            detailed_individually=True,
            )
        def member_query(self, _api, cg):
            set_renderer_query_fields()
            return Dataset.iter_all(_api,
                filter=_api.create_filter().member_of("cg_id", *[c.id for c in cg]),
                hook=partial(self._member_query_hook, pool_names_by_id=LazyIdNameMapping(Pool, _api)))

        def _member_query_hook(self, volumes, pool_names_by_id=None):
            has_rmr_fields = None
            for v in volumes:
                if pool_names_by_id:
                    v.pool = pool_names_by_id[v.pool_id]
                v.thin = v.provtype == "THIN"
                if has_rmr_fields is None:
                    has_rmr_fields = "rmr_source" in v and "rmr_target" in v
                if has_rmr_fields:
                    v.replication_role = ReplicaRoleOrNone.SOURCE if v.rmr_source \
                                         else ReplicaRoleOrNone.TARGET if v.rmr_target \
                                         else None
            return volumes

        @command(
            description="Rename {}".format(CgType.DESCRIPTION_SUBJECT),
            types=(CgType, Name),
            substitute=dict(
                cg="name",
                ),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def rename(self, _api, cg, new_name):
            return self._rename(_api, cg, new_name)

    CommonCgCommandsMixin.__name__ += "__" + CgType.__name__
    return CommonCgCommandsMixin

@commandgroup(prefix="cg")
class ConsistencyGroupCommands(ConsistencyGroupCommandsBase,
                               create_common_cg_commands_mixin(ConsistencyGroup, MasterVolume),
                               ReplicationCommandsMixin):
    ENTITY_TYPE = ConsistencyGroup
    QUERY_FIELDS = strlist("name pool members_count is_replicated")
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist("created_at updated_at")

    @command(
        description="List existing {}s".format(ConsistencyGroup.DESCRIPTION_SUBJECT),
        types=(ConsistencyGroup.List, Pool, MasterVolume),
        substitute=dict(
            cgs="name",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=ConsistencyGroup,
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, cgs=None, pool=None, member=None):
        if member and not member.cg_id:
            return ()
        return self._common_query(_api, cgs, pool_id=pool, id=member.cg_id if member else None)

    @command(
        description="Create a new consistency group",
        types=(Name, Pool),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, pool=None):
        pool = get_target_pool(_api, pool)
        return self._create(_api,
            name=name,
            pool_id=pool.id,
            )

    @command(
        description="Restore data state of consistency group from snapshot group",
        types=(ConsistencyGroup, SnapGroup),
        substitute=dict(
            cg="name",
            sg="source",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def restore(self, _api, cg, sg):
        self._post(_api, cg, "restore", data=dict(
            source_id=sg.id
            ))
        return self._result_message(cg, 'restored from snapshot group "{}"'.format(sg.name))

    @command(
        description='Add member to consistency group',
        types=(ConsistencyGroup, MasterVolume.List, RemoteEntity.List, Name.List),
        substitute=dict(
            cg='name',
            members='member',
            targets='target',
            new_target_names='new_target_name',
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add_member(self, _api, cg, members, targets=None, new_target_names=None):
        def type_and_subject(member):
            return '%s "%s"' % (self._get_member_type_name(member).capitalize(),
                                member.name)

        replicated_count = len([x for x in members if x.rmr_target or x.rmr_source])
        if replicated_count == len(members) or not cg.is_replicated:
            if targets or new_target_names:
                error = ('%s is already replicated' % type_and_subject(members[0])
                         if replicated_count else '%s is not replicated' %
                         self._result_subject_and_entity(cg, capitalize=True))
                raise BadArguments('%s, target and new_target_name are '
                                   'not applicable' % error)

            with _api.aggregation() as aggregator:
                for member in members:
                    aggregator.next_task(type_and_subject(member), 'Added')
                    self._add_member_simple(_api, cg, member)

        else:
            if replicated_count not in (0, len(members)):
                raise BadArguments('Either all members should be replicated, '
                                   'or none of them')
            if targets and new_target_names:
                raise BadArguments('Cannot specify both target and '
                                   'new_target_name at the same time')
            if targets:
                if len(targets) != len(members):
                    raise BadArguments('Please specify an equal number of '
                                       'members and targets')
                new_target_names = [None] * len(members)
            elif new_target_names:
                if len(new_target_names) != len(members):
                    raise BadArguments('Please specify an equal number of '
                                       'members and new_target_name items')
                targets = [None] * len(members)
            else:
                targets = new_target_names = [None] * len(members)

            link = self._get_link_from_cg(_api, cg)
            with get_remote_system(_api, link).api as remote_api:
                with _api.aggregation() as aggregator:
                    for member, target, new_target_name in zip(members, targets, new_target_names):
                        aggregator.next_task(type_and_subject(member), 'Added')
                        self._add_member_and_replicate(_api, remote_api, cg, member, target, new_target_name)

        if len(members) == 1:
            return ('%s added to consistency group "%s"' %
                    (type_and_subject(members[0]), cg.name))
        return '%d members added to consistency group "%s"' % (len(members), cg.name)

    def _add_member_simple(self, _api, cg, member):
        return self._post(_api, cg, 'members', data=dict(dataset_id=member.id))

    def _add_member_and_replicate(self, _api, remote_api, cg, member, target, new_target_name):
        pair = Munch(
            local_entity_id=member.id,
            local_base_action='NO_BASE_DATA',
            )
        if target:
            pair.update(
                remote_entity_id=Dataset.get_by_name(remote_api, target).id if target else None,
                remote_base_action='NO_BASE_DATA',
                )
        else:
            pair.update(
                remote_entity_name=new_target_name,
                remote_base_action='CREATE',
                )
        return self._post(_api, cg, 'members', data=Munch(
            dataset_id=member.id,
            replication_pair_info=pair,
            ))

    def _get_link_from_cg(self, _api, cg):
        replicas = Replica.get_all(_api, filter=_api.create_filter(local_cg_id=cg.id))
        if len(replicas) != 1:
            raise CommandFailed("Replica for {} not found".format(self._result_subject_and_entity(cg)) if not replicas
                                else "{} is replicated to multiple systems".format(self._result_subject_and_entity(cg), capitalize=True))
        return ReplicationLink.get_by_id(_api, replicas[0].link_id)

    @command(
        description="Remove member from consistency group",
        types=(ConsistencyGroup, MasterVolume, YesNo, YesNo, YesNo, Name),
        substitute=dict(
            cg="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove_member(self, _api, cg, member, retain_staging_area=False, force=False, create_new_replica=False, new_replica_name=None):
        if not cg.is_replicated:
            if create_new_replica or new_replica_name:
                raise BadArguments, "The consistancy group is not replicated, create_new_replica and new_replica_name parameters are not needed"
            self._del(_api, cg, "members", member.id)
            return '{} "{}" removed from consistency group "{}"'.format(self._get_member_type_name(member).capitalize(), member.name, cg.name)
        link = self._get_link_from_cg(_api, cg)
        if force and create_new_replica:
            raise BadArguments, "Cannot use force parameter with create_new_replica parameter"
        params = dict(
            retain_staging_area=retain_staging_area,
            create_replica=create_new_replica,
            replica_name=new_replica_name,
            )
        send_request = partial(self._del, _api, cg, "members", member.id, params=params)
        # result = None
        try:
            with get_remote_system(_api, link).api:
                # result = send_request()
                send_request()
        except ManagementException as e:
            if not self._should_retry_forcefully(e):
                raise
            elif not force:
                if not yesno("{}{}Would you like to perform the operation forcefully?".format(e.message,"\n" if e.message else ""), always_prompt=True):
                    raise CommandAborted
            params.update(
                force_if_no_remote_credentials="true",
                force_if_remote_error="true",
                force_on_target="true",
                )
            # result = send_request(approved=True)
            send_request(approved=True)
        staging_message = "and its staging area " + ("retained" if retain_staging_area else "discarded")
        # if retain_staging_area and result and result.entity_pairs:
        #     non_reclaimed_names = sorted(pair.local_entity_name
        #                                  for pair in result.entity_pairs
        #                                  if not pair.get("_local_reclaimed_snapshot_id", False))
        #     if non_reclaimed_names:
        #         staging_message = "but there was nothing in the staging area to retain" if len(result.entity_pairs) == len(non_reclaimed_names) \
        #                           else "and its staging area was retained, except for the following datasets: " + humanly_join(non_reclaimed_names)
        return '{} "{}" removed from consistency group "{}", '.format(self._get_member_type_name(member).capitalize(), member.name, cg.name) + staging_message


    @command(
        description="Move consistency group to a different pool",
        types=(ConsistencyGroup, Pool, YesNo),
        substitute=dict(
            cg="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def move(self, _api, cg, pool, move_capacity=False):
        if cg.pool_id == pool.id:
            return self._result_message(cg, 'already in {} "{}"'.format(Pool.get_entity_name(), pool.name))
        self._post(_api, cg, "move", data=dict(
            pool_id=pool.id,
            with_capacity=move_capacity,
            ))
        return self._result_message(cg, 'moved to pool "{}" {} its capacity'
                                        .format(pool.name, "along with" if move_capacity else "without"))

@commandgroup(prefix="snap", parent=ConsistencyGroupCommands)
class SnapshotGroupCommands(ConsistencyGroupCommandsBase,
                            create_common_cg_commands_mixin(SnapGroup, VolumeSnap)):
    ENTITY_TYPE = SnapGroup
    QUERY_FIELDS = strlist("name pool members_count is_replicated")
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist("created_at updated_at")

    @command(
        description="List existing {}s".format(SnapGroup.DESCRIPTION_SUBJECT),
        types=(SnapGroup.List, ConsistencyGroup, Pool, VolumeSnap),
        substitute=dict(
            snapgroups="name",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=SnapGroup,
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, snapgroups=None, source=None, pool=None, member=None):
        if member and not member.cg_id:
            return ()
        return self._common_query(_api, snapgroups, parent_id=source, pool_id=pool, id=member.cg_id if member else None)

    @command(
        description="Create a snapshot of a consistency group",
        types=(ConsistencyGroup, Name, Name, Name),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, source, name, snap_prefix=None, snap_suffix=None):
        if not snap_suffix and not snap_prefix:
            raise BadArguments("Either snap_prefix, snap_suffix, or both must be specified")
        return self._create(_api,
            name=name,
            parent_id=source,
            snap_prefix=snap_prefix,
            snap_suffix=snap_suffix,
            )
