from munch import Munch

from infi.shell import render
from infi.shell.commands import commandgroup, command
from infi.shell.types import Str
from infi.shell.utils import strlist

from .options import QueryOptions
from .types import ObjType, Name, Metadata
from . import EntityCommands, roles, render_collection, set_renderer_query_fields

from ..pre_v2_patches import pre_v2_command, v2_command

@commandgroup(prefix="metadata")
class MetadataCommands(EntityCommands):
    ENTITY_TYPE = object()  # Dummy

    def _get_object(self, _api, objtype, objname):
        return objtype.get_by_name(_api, objname)

    @command(
        description="Set metadata on object",
        types=(ObjType, Name, Str, Str),
        substitute=dict(objname="object"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set(self, _api, objtype, objname, key, value):
        obj = self._get_object(_api, objtype, objname)
        data = {key : value}
        _api.put("metadata/{}".format(obj.id), data)
        return self._result_message(obj, 'metadata key "{}" set to "{}"'.format(key, value), subject=objtype)


    @pre_v2_command(
        name="get",
        description="Get metadata from object",
        types=(ObjType, Name, Str),
        substitute=dict(objname="object"),
        )
    @render(
        order=["key"],
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def pre_v2_get(self, _api, objtype, objname, key=None):
        obj = self._get_object(_api, objtype, objname)
        result = None
        if key is None:
            result = [Munch((("key", k), ("value", v)))
                      for k, v in _api.get("metadata/{}".format(obj.id)).iteritems()]
        else:
            try:
                result = [Munch(key=key, value=_api.get(u"metadata/{}/{}".format(obj.id, key)))]
            except NotFound:
                raise NoSuchKey(key)
        return sorted(result, key=lambda key_value_object: key_value_object.key)


    @v2_command(
        description="Get metadata from object",
        types=(ObjType, Name, Str),
        substitute=dict(objname="object"),
        options=QueryOptions(fields=strlist("key value"), sortable=strlist("key"), detailed=True),
        )
    @render_collection(
        entity=Metadata,
        fields=strlist("key value"),
        detailed_fields=strlist("key value"),
        detailed_individually=True,
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def get(self, _api, objtype, objname, key=None):
        set_renderer_query_fields()
        obj = self._get_object(_api, objtype, objname)
        if key is None:
            return _api.paged_query("metadata/{}".format(obj.id),
                params=_api.create_sort().asc("key").to_params(),
                hook=self._query_hook,
                )
        return self._query_hook([_api.get(u"metadata/{}/{}".format(obj.id, key))])

    def _query_hook(self, metadata_objects):
        for m in metadata_objects:
            yield Munch(key=m.key, value=m.value)

    @command(
        description="Remove a single metadata key from object",
        types=(ObjType, Name, Str),
        substitute=dict(objname="object"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove(self, _api, objtype, objname, key):
        obj = self._get_object(_api, objtype, objname)
        _api.delete(u"metadata/{}/{}".format(obj.id, key))
        return self._result_message(obj, 'metadata key "{}" removed'.format(key), subject=objtype)

    @command(
        description="Clear all metadata keys from object",
        types=(ObjType, Name),
        substitute=dict(objname="object"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear(self, _api, objtype, objname):
        obj = self._get_object(_api, objtype, objname)
        _api.delete("metadata/{}".format(obj.id))
        return self._result_message(obj, "metadata cleared", subject=objtype)
