from functools import partial
from collections import defaultdict
from munch import Munch

from infi.shell.exceptions import BadArguments, BadValue
from infi.shell.commands import commandgroup, command
from infi.shell.types import YesNo
from infi.shell.utils import strlist

from ..exceptions import ManagementException
from ..renderer import render_capacity
from .hierarchical_base import HierarchicalEntityCommandsBase
from .types import (
    Pool,
    Volume,
    VolumeSnap,
    VolumeClone,
    VolumeOrClone,
    VolumeSnapOrClone,
    MasterVolume,
    ConsistencyGroup,
    SnapGroup,
    Cluster,
    Host,
    Lun,
    Name,
    Size,
    SizeDelta,
    CapacityUnit,
    VolumeSeries,
    ReplicaRoleOrNone,
    )
from .options import QueryOptions
from .mapping import MappingCommandsBase
from .utils import get_target_pool
from . import render_collection, roles, set_rendered_unit, set_renderer_query_fields

from ..pre_v2_patches import is_pre_v2_system

VOLUME_QUERY_FIELDS           = strlist("name             thin size used allocated tree_allocated pool write_protected mapped")
DETAILED_VOLUME_QUERY_FIELDS  = strlist("name      serial thin size used allocated tree_allocated pool write_protected mapped ssd_enabled replication_role created_at updated_at")

SNAP_QUERY_FIELDS             = strlist("name             thin size used allocated                pool                 mapped                              created_at")
DETAILED_SNAP_QUERY_FIELDS    = strlist("name      serial thin size used allocated                pool                 mapped ssd_enabled replication_role created_at updated_at")

CLONE_QUERY_FIELDS            = strlist("name             thin size used allocated tree_allocated pool write_protected mapped                              created_at")
DETAILED_CLONE_QUERY_FIELDS   = DETAILED_VOLUME_QUERY_FIELDS

VOLUME_TREE_FIELDS            = strlist("name type             size used allocated                pool write_protected                                     created_at")
DETAILED_VOLUME_TREE_FIELDS   = strlist("name type serial thin size used allocated tree_allocated pool write_protected mapped ssd_enabled replication_role created_at updated_at")

MAP_QUERY_FIELDS = strlist("name target_type target_name luns")

SORTABLE_FIELDS = strlist("name serial size write_protected mapped ssd_enabled created_at updated_at")

QUERY_SUBSTITUTES = dict(
    write_protected="wp",
    ssd_enabled="ssd_cache",
    )

DETAILED_QUERY_SUBSTITUTES = dict(
    write_protected="write_protected",
    )

class VolumeCommandsBase(HierarchicalEntityCommandsBase):
    HIERARCHICAL_TYPE = Volume
    VOLUME_TYPE_NAMES = Munch(
        MASTER="volume",
        SNAP="snapshot",
        CLONE="clone",
        )

    def _query_hook(self, volumes, pool_names_by_id=None):
        has_rmr_fields = None
        for v in volumes:
            if pool_names_by_id:
                v.pool = pool_names_by_id[v.pool_id]
            v.type = self._get_volume_type(v)
            v.thin = v.provtype == self.THIN
            if has_rmr_fields is None:
                has_rmr_fields = "rmr_source" in v and "rmr_target" in v
            if has_rmr_fields:
                v.replication_role = ReplicaRoleOrNone.SOURCE if v.rmr_source \
                                     else ReplicaRoleOrNone.TARGET if v.rmr_target \
                                     else None
        return volumes

    def _get_volume_type(self, volume):
        if volume.type == "MASTER":
            return "VOLUME"
        return volume.type

    def _get_volume_type_name(self, volume):
        return self.VOLUME_TYPE_NAMES.get(volume.type) or "volume"

def create_common_volume_commands_mixin(VolumeType):

    class CommonVolumeCommandsMixin(MappingCommandsBase):

        @command(
            description="Rename {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, Name),
            substitute=dict(volume="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def rename(self, _api, volume, new_name):
            return self._rename(_api, volume, new_name)

        @command(
            description="Delete {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType.List,),
            substitute=dict(volume="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def delete(self, _api, volume):
            return self._delete(_api, volume)

        @command(
            description="Map {} to host or cluster".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType.List, Host, Cluster, Lun),
            substitute=dict(volumes="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def map(self, _api, volumes, host=None, cluster=None, lun=None):
            if host and cluster or not (host or cluster):
                raise BadArguments("Either host or cluster must be specified")
            if lun and len(volumes) != 1:
                raise BadArguments("LUN may be specified only when mapping a single " + VolumeType.get_entity_name())
            mappable_type, mappable = self._get_mappable(host, cluster)
            return self._map(_api, mappable_type, mappable, volumes, lun)

        @command(
            description="Unmap {} from host or cluster".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType.List, Host, Cluster),
            substitute=dict(volumes="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def unmap(self, _api, volumes, host=None, cluster=None):
            if not host and not cluster:
                raise BadArguments("Either host or cluster must be specified")
            mappable_type, mappable = self._get_mappable(host, cluster)
            return self._unmap(_api, mappable_type, mappable, volumes)

        def _get_mappable(self, host, cluster):
            if host:
                return Host, host
            if cluster:
                return Cluster, cluster
            raise AssertionError("Undefined mappable type")

        @command(
            description="List host and cluster mappings by volume",
            types=(VolumeType.List, Host, Cluster),
            substitute=dict(volume="name"),
            options=QueryOptions(fields=MAP_QUERY_FIELDS),
            )
        @render_collection(
            entity=Volume,
            fields=MAP_QUERY_FIELDS,
            substitute=dict(
                luns="lun_id",
            ),
            )
        def map_query(self, _api, volume=None, host=None, cluster=None):
            if host and cluster:
                raise BadArguments("Either host, cluster, or neither may be specified")
            show_all_mappables = all(mappable is None for mappable in (host, cluster))
            def get_mappables(mappable_type, specified_mappable):
                return mappable_type.iter_all(_api) if show_all_mappables else (specified_mappable,) if specified_mappable else ()
            mappings = defaultdict(partial(defaultdict, partial(defaultdict, Munch)))
            mapping_infos = defaultdict(list)
            for mappable_type, specified_mappable, lun_predicate in (
                (Host,       host,      lambda lun: not lun.clustered),
                (Cluster,    cluster,   None),
                ):
                for mappable in get_mappables(mappable_type, specified_mappable):
                    for lun in mappable.luns:
                        if not lun_predicate or lun_predicate(lun):
                            mapping_info = mappings[lun.volume_id][mappable_type][mappable.id]
                            if not mapping_info:
                                mapping_info.target_name = mappable.name
                                mapping_info.target_type = mappable_type.get_entity_name().upper()
                                mapping_info.luns = []
                                mapping_infos[lun.volume_id].append(mapping_info)
                            mapping_info.luns.append(lun.lun)
            return self._query(_api, volume, _hook=partial(self._map_query_hook, mapping_infos=mapping_infos))

        def _map_query_hook(self, volumes, mapping_infos):
            for v in volumes:
                for info in sorted(mapping_infos.get(v.id, ()), key=lambda i: i.target_name):
                    info.name = v.name
                    info.luns.sort()
                    yield info

        @command(
            description="Enable or disable SSD cache",
            types=(VolumeType, YesNo, YesNo),
            substitute=dict(volume="name", enabled="ssd"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def cache(self, _api, volume, enabled, recursive=False):
            return self._update_recursive(_api,
                entity=volume,
                recurse=recursive,
                success_msg="{} cached on SSD".format("now" if enabled else "no longer"),
                success_msg_with_is=True,
                noop_msg="already{} cached on SSD".format("" if enabled else " not"),
                noop_msg_with_is=True,
                props=dict(
                    ssd_enabled=enabled,
                    ),
                predicate=lambda dataset: dataset.ssd_enabled != enabled,
                )

    return CommonVolumeCommandsMixin

def create_writable_volume_commands_mixin(VolumeType):

    class WritableVolumeCommandsMixin(object):

        @command(
            description="Resize {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, SizeDelta),
            substitute=dict(volume="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def resize(self, _api, volume, size):
            size = size.compute_size(volume.size)
            if volume.size == size:
                return self._result_message(volume, "size is already {}".format(render_capacity(volume.size)))
            self._update(_api, volume, size=size)
            return self._result_message(volume, "resized")

        @command(
            description="Restore data state of {} from previous snapshot".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, VolumeSnapOrClone),
            substitute=dict(volume="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def restore(self, _api, volume, source):
            if volume.id == source.id:
                raise BadValue('"{}" was specified as both arguments'.format(volume.name))
            data = dict(
                source_id=source.id
                )
            if is_pre_v2_system():
                data = source.id
            self._post(_api, volume, "restore", data=data)
            return self._result_message(volume, 'restored from {} "{}"'.format(self._get_volume_type_name(source), source.name))

        @command(
            description="Disable write access on {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, YesNo),
            substitute=dict(volume="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_protect(self, _api, volume, recursive=False):
            return self._update_recursive(_api,
                entity=volume,
                recurse=recursive,
                success_msg="now write-protected",
                success_msg_with_is=True,
                noop_msg="already write-protected",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=True,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and not dataset.write_protected,
                )

        @command(
            description="Enable write access on {}".format(VolumeType.DESCRIPTION_SUBJECT),
            types=(VolumeType, YesNo),
            substitute=dict(volume="name"),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def write_enable(self, _api, volume, recursive=False):
            return self._update_recursive(_api,
                entity=volume,
                recurse=recursive,
                success_msg="now write-enabled",
                success_msg_with_is=True,
                noop_msg="already write-enabled",
                noop_msg_with_is=True,
                props=dict(
                    write_protected=False,
                    ),
                predicate=lambda dataset: dataset.type != "SNAP" and dataset.write_protected,
                )

    return WritableVolumeCommandsMixin

def create_child_volume_commands_mixin(VolumeType, ParentType, parent_arg, create_parent_arg):

    class ChildVolumeCommandsMixin(object):

        @command(
            description="Create a new {} from {}".format(VolumeType.DESCRIPTION_SUBJECT, ParentType.DESCRIPTION_SUBJECT),
            types=(ParentType, Name),
            substitute=dict(parent=create_parent_arg),
            )
        @roles.INFINIDAT.ADMIN.POOLADMIN
        def create(self, _api, parent, name):
            parent = self._create_and_return_entity(_api, parent_id=parent.id, name=name)
            return self._result_message(parent, "created", subject=VolumeType.get_entity_name())

        @command(
            description="Display hierarchy of snapshots and clones",
            types=(VolumeType, CapacityUnit),
            substitute=dict(volume="name"),
            options=QueryOptions(fields=DETAILED_VOLUME_TREE_FIELDS, sortable=VOLUME_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
            )
        @render_collection(
            entity=Volume,
            fields=VOLUME_TREE_FIELDS,
            substitute=QUERY_SUBSTITUTES,
            )
        def tree(self, _api, volume, unit=None):
            set_rendered_unit(unit)
            set_renderer_query_fields()
            return self._get_tree(_api, volume)

    return ChildVolumeCommandsMixin

@commandgroup(prefix="vol")
class VolumeCommands(VolumeCommandsBase,
                     create_common_volume_commands_mixin(MasterVolume),
                     create_writable_volume_commands_mixin(MasterVolume)):
    ENTITY_TYPE = MasterVolume
    @command(
        description="List existing volumes",
        types=(MasterVolume.List, ConsistencyGroup, Pool, CapacityUnit, YesNo, YesNo, YesNo),
        substitute=dict(volume="name"),
        options=QueryOptions(fields=DETAILED_VOLUME_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=VOLUME_QUERY_FIELDS,
        detailed_fields=DETAILED_VOLUME_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, volume=None, cg=None, pool=None, unit=None, mapped=None, write_protected=None, thin=None):
        set_rendered_unit(unit)
        provtype = None
        if thin is not None:
            provtype = self.THIN if thin else self.THICK
        return self._common_query(_api, volume, cg_id=cg, pool_id=pool, mapped=mapped, write_protected=write_protected, provtype=provtype)

    @command(
        description="Create a new volume",
        types=(Name.List, Size, Pool, YesNo, YesNo, VolumeSeries),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, size, pool=None, thin=False, ssd_cache=None, series=None):
        pool = get_target_pool(_api, pool)
        if series:
            # In case of multiple volumes, since this is not an atomic
            # operation, ensure there's enough free space for all requested
            # volumes.
            total_size = size * series.total
            if total_size > pool.free_virtual_space:
                raise ManagementException(409, Munch(code="POOL_VIRTUAL_CAPACITY", message="Pool virtual capacity reached."))
            if total_size > pool.free_physical_space:
                raise ManagementException(409, Munch(code="POOL_PHYSICAL_CAPACITY", message="Pool physical capacity reached."))
        return self._create_single_or_multiple_or_series(_api, name, series,
            size=size,
            pool_id=pool.id,
            provtype=self._get_provtype(thin),
            ssd_enabled=ssd_cache,
            )

    @command(
        description="Move volume to a different pool",
        types=(MasterVolume, Pool, YesNo),
        substitute=dict(volume="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def move(self, _api, volume, pool, move_capacity=False):
        if volume.pool_id == pool.id:
            return self._result_message(volume, 'already in {} "{}"'.format(Pool.get_entity_name(), pool.name))
        _api.post(self.ENTITY_TYPE.get_uri(volume.id, "move"), dict(
            pool_id=pool.id,
            with_capacity=move_capacity,
            ))
        return self._result_message(volume, 'moved to pool "{}" {} its capacity'
                                            .format(pool.name, "along with" if move_capacity else "without"))

    @command(
        description="Set volume provisioning type",
        types=(MasterVolume, YesNo),
        substitute=dict(volume="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_thin(self, _api, volume, thin):
        prov_name = "thin" if thin else "thick"
        if volume.provtype.upper() == self._get_provtype(thin):
            return self._result_message(volume, "already set to {} provisioning".format(prov_name))
        self._update(_api, volume, provtype=self._get_provtype(thin))
        return self._result_message(volume, "set to {} provisioning".format(prov_name))

    @command(
        description="Estimate reclaimed capacity (deletion simulation)",
        types=(Volume,),
        substitute=dict(volume="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def reclaimable(self, _api, volume):
        data = dict(entities=[volume.id])
        result = _api.post(Volume.get_uri("delete_simulation"), data)
        return 'Estimated reclaimable space for volume "{}" and its tree descendants: {}'.format(volume.name, render_capacity(result.space_reclaimable))

    @command(
        description="Display hierarchy of volumes, snapshots and clones",
        types=(MasterVolume, Pool, CapacityUnit),
        substitute=dict(volume="name"),
        options=QueryOptions(fields=DETAILED_VOLUME_TREE_FIELDS, sortable=VOLUME_TREE_FIELDS, substitute=DETAILED_QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Volume,
        fields=VOLUME_TREE_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def tree(self, _api, volume=None, pool=None, unit=None):
        set_rendered_unit(unit)
        set_renderer_query_fields()
        return self._get_tree(_api, volume, pool)

@commandgroup(prefix="snap", parent=VolumeCommands)
class VolumeSnapshotCommands(VolumeCommandsBase,
                             create_common_volume_commands_mixin(VolumeSnap),
                             create_child_volume_commands_mixin(
                                 VolumeType=VolumeSnap,
                                 ParentType=VolumeOrClone,
                                 parent_arg="vol",
                                 create_parent_arg="source",
                                 )):
    ENTITY_TYPE = VolumeSnap

    @command(
        description="List existing snapshots",
        types=(VolumeSnap.List, VolumeOrClone, SnapGroup, Pool, CapacityUnit, YesNo),
        substitute=dict(snap="name"),
        options=QueryOptions(fields=DETAILED_SNAP_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=SNAP_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_SNAP_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        )
    def query(self, _api, snap=None, source=None, sg=None, pool=None, unit=None, mapped=None):
        set_rendered_unit(unit)
        return self._common_query(_api, snap, parent_id=source, cg_id=sg, pool_id=pool, type=VolumeSnap.ENTITY_MGMT_SUBTYPE, mapped=mapped)

    @command(
        description="Create a consistent set of snapshots of multiple volumes or clones",
        types=(VolumeOrClone.List, Name, Name),
        )
    @roles.INFINIDAT
    def group_create(self, _api, source, snap_prefix=None, snap_suffix=None):
        if not snap_suffix and not snap_prefix:
            raise BadArguments("Either snap_prefix, snap_suffix, or both must be specified")
        _api.post(VolumeOrClone.get_uri("group_snapshot"), dict(
            entities=[dict(id=entity.id) for entity in source],
            snap_prefix=snap_prefix,
            snap_suffix=snap_suffix,
            ))
        return "Consistent snapshots created"

@commandgroup(prefix="clone", parent=VolumeCommands)
class VolumeCloneCommands(VolumeCommandsBase,
                          create_common_volume_commands_mixin(VolumeClone),
                          create_writable_volume_commands_mixin(VolumeClone),
                          create_child_volume_commands_mixin(
                              VolumeType=VolumeClone,
                              ParentType=VolumeSnap,
                              parent_arg="snap",
                              create_parent_arg="snap",
                              )):
    ENTITY_TYPE = VolumeClone

    @command(
        description="List existing clones",
        types=(VolumeClone.List, VolumeSnap, ConsistencyGroup, Pool, CapacityUnit, YesNo, YesNo),
        substitute=dict(clone="name"),
        options=QueryOptions(fields=DETAILED_CLONE_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Volume,
        fields=CLONE_QUERY_FIELDS,
        detailed_fields=DETAILED_CLONE_QUERY_FIELDS,
        detailed_substitute=DETAILED_QUERY_SUBSTITUTES,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, clone=None, snap=None, cg=None, pool=None, unit=None, mapped=None, write_protected=None):
        set_rendered_unit(unit)
        return self._common_query(_api, clone, parent_id=snap, cg_id=cg, pool_id=pool, type=VolumeClone.ENTITY_MGMT_SUBTYPE, mapped=mapped, write_protected=write_protected)
