from functools import partial
from munch import Munch

from infi.shell import commandgroup
from infi.shell.types import Str, YesNo, Int
from infi.shell.exceptions import BadArguments, BadValue
from infi.shell.utils import strlist

from .filesystem import command, FilesystemCommands, FilesystemCommandsMixin
from .types import (
    Size,
    Export,
    ExportClient,
    ExportAccessLevel,
    Filesystem,
    )
from .options import QueryOptions
from . import (
    EntityCommands,
    roles,
    render_collection,
    LazyEntityMapping,
    will_render_any_field,
    )

class FilesystemExportCommandsBase(EntityCommands):

    def _get_lazy_fs_mapping(self, api, ids=None, preloaded=None):
        filter = None if ids is None else api.create_filter().member_of("id", *ids)
        return LazyEntityMapping(Filesystem, api, fields=("id", "name", "type"), filter=filter, preloaded_entities=preloaded)

@commandgroup(parent=FilesystemCommands, prefix="export")
class FilesystemExportCommands(FilesystemCommandsMixin, FilesystemExportCommandsBase):
    ENTITY_TYPE = Export
    QUERY_FIELDS = strlist("""
        export_path
        inner_path
        filesystem
        filesystem_type
        enabled
        """)
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist("""
        acl
        anonymous_gid
        anonymous_uid
        make_all_users_anonymous
        max_write
        max_read
        pref_write
        pref_read
        pref_readdir
        privileged_port
        """)
    QUERY_SUBSTITUTES = dict(
        inner_path="internal_path",
        filesystem_type="type",
        anonymous_gid="anonymous_gid",
        anonymous_uid="anonymous_uid",
        make_all_users_anonymous="squash_all_users",
        max_write="maximum_write_size",
        max_read="maximum_read_size",
        pref_write="preferred_write_size",
        pref_read="preferred_read_size",
        pref_readdir="preferred_readdir_size",
        privileged_port="privileged_ports_only",
        )
    SORTABLE_FIELDS = DETAILED_QUERY_FIELDS

    @command(
        description="List existing filesystem exports",
        types=(Export.List, Str.List, Filesystem.List),
        substitute=dict(
            export="export_path",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Export,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, export=None, internal_path=None, fs=None):
        filesystem_by_id = self._get_lazy_fs_mapping(_api, preloaded=fs) if fs else None
        return self._query(_api, export,
            inner_path=internal_path,
            filesystem_id=fs,
            _hook=partial(self._query_hook,
                api=_api,
                filesystem_by_id=filesystem_by_id,
                with_fs_info=will_render_any_field("filesystem", "filesystem_type"),
                ),
            )

    def _query_hook(self, exports, api=None, filesystem_by_id=None, with_fs_info=False):
        if with_fs_info:
            if not filesystem_by_id and api:
                filesystem_by_id = self._get_lazy_fs_mapping(api, ids=(e.filesystem_id for e in exports))
            if filesystem_by_id:
                for e in exports:
                    fs = filesystem_by_id[e.filesystem_id]
                    e.update(
                        filesystem=fs.name,
                        filesystem_type=self._get_fs_type(fs),
                        )
        return exports

    def _generate_export_name(self, export_path):
        return export_path.strip("/").replace("/", "_")

    @command(
        description="Create a new filesystem export",
        types=(Filesystem, Str, Str, Export,
               YesNo, Int, Int,
               Size, Size,
               Size, Size, Size,
               YesNo),
        substitute=dict(
            permissions_export="copy_permissions_from",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, fs, export_path, internal_path=None, permissions_export=None,
                           squash_all_users=None, anonymous_gid=None, anonymous_uid=None,
                           maximum_write_size=None, maximum_read_size=None,
                           preferred_write_size=None, preferred_read_size=None, preferred_readdir_size=None,
                           privileged_port=None):
        props = self._create_export_props(**locals())
        props.update(
            filesystem_id=fs.id,
            export_path=export_path,
            inner_path=internal_path,
            )
        return self._create(_api, **props)

    def _create_export_props(_self, permissions=None, permissions_export=None,
                                    squash_all_users=None, anonymous_gid=None, anonymous_uid=None,
                                    maximum_write_size=None, maximum_read_size=None,
                                    preferred_write_size=None, preferred_read_size=None, preferred_readdir_size=None,
                                    privileged_port=None,
                                    **__):
        if permissions and permissions_export:
            raise BadArguments("Please specify either permissions or copy_permissions_from")
        props = Munch(
            make_all_users_anonymous=squash_all_users,
            anonymous_gid=anonymous_gid,
            anonymous_uid=anonymous_uid,
            max_write=maximum_write_size,
            max_read=maximum_read_size,
            pref_write=preferred_write_size,
            pref_read=preferred_read_size,
            pref_readdir=preferred_readdir_size,
            privileged_port=privileged_port,
            )
        if permissions:
            _self._apply_permissions_on_export_props(props, permissions)
        elif permissions_export:
            props.permissions = permissions_export.permissions
        return props

    def _apply_permissions_on_export_props(self, props, permissions):
        raise NotImplementedError

    @command(
        description="Modify existing filesystem export",
        types=(Export, Export,
               YesNo, Int, Int,
               Size, Size,
               Size, Size, Size,
               YesNo),
        substitute=dict(
            export="export_path",
            permissions_export="copy_permissions_from",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def modify(self, _api, export, permissions_export=None,
                           squash_all_users=None, anonymous_gid=None, anonymous_uid=None,
                           maximum_write_size=None, maximum_read_size=None,
                           preferred_write_size=None, preferred_read_size=None, preferred_readdir_size=None,
                           privileged_port=None):
        props = self._create_export_props(**locals())
        if not any(value is not None for value in props.itervalues()):
            raise BadArguments("At least one more argument must be specified")
        self._update(_api, export, **props)
        return self._result_message(export, "modified")

    @command(
        description="Enable filesystem export",
        types=(Export,),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def enable(self, _api, export):
        if export.enabled:
            return self._result_message(export, "is already enabled")
        self._update(_api, export, enabled=True)
        return self._result_message(export, "enabled")

    @command(
        description="Disable filesystem export",
        types=(Export,),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def disable(self, _api, export):
        if not export.enabled:
            return self._result_message(export, "is already disabled")
        self._update(_api, export, enabled=False)
        return self._result_message(export, "disabled")

    @command(
        description="Delete filesystem export",
        types=(Export.List,),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, export):
        return self._delete(_api, export)

@commandgroup(parent=FilesystemExportCommands, prefix="permission")
class FilesystemExportPermissionCommands(FilesystemCommandsMixin, FilesystemExportCommandsBase):
    ENTITY_TYPE = Export
    QUERY_FIELDS = strlist("""
        export_path
        client
        access
        """)
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist("""
        filesystem
        filesystem_type
        """)
    QUERY_SUBSTITUTES = dict(
        access="access_level",
        filesystem_type="type",
        )
    SORTABLE_FIELDS = strlist("export_path")

    @command(
        description="List permissions for existing filesystem exports",
        types=(Export.List, ExportClient.List, ExportAccessLevel.List, Filesystem.List),
        substitute=dict(
            exports="export_path",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Export,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, exports=None, client=None, access=None, fs=None):
        filesystem_by_id = self._get_lazy_fs_mapping(_api, preloaded=fs) if fs else None
        return self._query(_api, exports,
            _hook=partial(self._query_hook,
                api=_api,
                client=client,
                access=access,
                filesystem_by_id=filesystem_by_id,
                with_fs_info=will_render_any_field("filesystem", "filesystem_type"),
                ),
            )

    def _query_hook(self, exports, api=None, client=None, access=None, filesystem_by_id=None, with_fs_info=False):
        if with_fs_info and api and not filesystem_by_id:
            filesystem_by_id = self._get_lazy_fs_mapping(api, ids=(e.filesystem_id for e in exports))
        permissions = (
            self._create_permissions_object(export, permission, client, access, filesystem_by_id)
            for export in exports
            for permission in sorted(export.permissions, key=lambda _p: _p.client)
            if permission.client
            )
        return self._filter_entities(permissions, client=client, access=access)

    def _create_permissions_object(self, export, permission, client, access, filesystem_by_id):
        obj = Munch(
            export_path=export.export_path,
            client=permission.client,
            access=ExportAccessLevel.from_mgmt_value(permission.access),
            )
        if filesystem_by_id:
            fs = filesystem_by_id[export.filesystem_id]
            obj.update(
                filesystem=fs.name,
                filesystem_type=self._get_fs_type(fs),
                )
        return obj

    @command(
        description="Add a new permission to an existing filesystem export",
        types=(Export, ExportClient, ExportAccessLevel),
        substitute=dict(
            export="export_path",
            ),
        )
    def add(self, _api, export, client, access):
        permissions = export.permissions
        for p in permissions:
            if p.client == client:
                raise BadValue('Client "{}" is already defined for {}, with {} access'
                               .format(client, self._result_subject_and_entity(export),
                                       ExportAccessLevel.from_mgmt_value(p.access)))
        permissions.append(dict(
            client=client,
            access=ExportAccessLevel.to_mgmt_value(access),
            ))
        self._update(_api, export, permissions=permissions)
        return '{} access granted to client "{}" on {}'.format(access, client, export.export_path)

    @command(
        description="Remove a permission from an existing filesystem export",
        types=(Export, ExportClient),
        substitute=dict(
            export="export_path",
            ),
        )
    def remove(self, _api, export, client):
        permissions = export.permissions
        removed_permission = None
        for i, p in enumerate(permissions):
            if p.client == client:
                removed_permission = p
                permissions.pop(i)
                break
        if removed_permission is None:
            raise BadValue('Client "{}" is not defined for {}'.format(client, self._result_subject_and_entity(export)))
        self._update(_api, export, permissions=permissions)
        return '{} access revoked from client "{}" on {}'.format(ExportAccessLevel.from_mgmt_value(p.access), client, export.export_path)

    @command(
        description="Add a new permission to an existing filesystem export",
        types=(Export, ExportClient, ExportAccessLevel),
        substitute=dict(
            export="export_path",
            ),
        )
    def modify(self, _api, export, client, access):
        permissions = export.permissions
        found = False
        for p in permissions:
            if p.client == client:
                found = True
                value = ExportAccessLevel.to_mgmt_value(access)
                if p.access == value:
                    raise BadValue('Client "{}" already has {} access to {}'.format(client, access, self._result_subject_and_entity(export)))
                p.access = value
        if not found:
            raise BadValue('Client "{}" is not defined for {}'.format(client, self._result_subject_and_entity(export)))
        permissions.append(dict(
            client=client,
            access=ExportAccessLevel.to_mgmt_value(access),
            ))
        self._update(_api, export, permissions=permissions)
        return '{} access granted to client "{}" on {}'.format(access, client, export.export_path)
