from munch import Munch
from itertools import chain
from functools import partial

from infi.shell import commandgroup, command
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments

from .types import Name, Host, Cluster, WWN, SCSIPortType, Volume, MasterVolume, VolumeSnap, VolumeClone, Lun
from .options import QueryOptions
from .mapping import MappingCommandsBase
from . import EntityCommands, render_collection, set_renderer_query_fields, roles, LazyIdNameMapping

class HostCommandsBase(EntityCommands, MappingCommandsBase):
    def _map(self, _api, mappable, lun, *volume_collections):
        all_volumes = list(chain(*((c or ()) for c in volume_collections)))
        if not len(all_volumes):
            raise BadArguments("At least one volume, snapshot, or clone must be specified")
        if lun is not None and len(all_volumes) != 1:
            raise BadArguments("Specifying LUN in conjunction with more than one volume, snapshot, or clone is not allowed")
        return super(HostCommandsBase, self)._map(_api, self.ENTITY_TYPE, mappable, all_volumes, lun)

    def _unmap(self, _api, mappable, lun, *volume_collections):
        all_volumes = list(chain(*((c or ()) for c in volume_collections)))
        if lun is None:
            if not len(all_volumes):
                raise BadArguments("At least one volume, snapshot, or clone must be specified")
        else:
            if len(all_volumes):
                raise BadArguments("Specifying LUN in conjunction with volumes, snapshots, or clones is not allowed")
            return self._unmap_lun(_api, mappable, lun)
        return super(HostCommandsBase, self)._unmap(_api, self.ENTITY_TYPE, mappable, all_volumes)

    def _map_query(self, _api, lun_container, volume=None):
        set_renderer_query_fields()
        name = lun_container.name if lun_container else None
        lun_containers = self._query(_api, name=name)
        if volume:
            return [Munch({self.ENTITY_TYPE.get_entity_name() : lc.name,
                           "lun" : mapping.lun,
                           "volume" : volume.name,
                           "target_type" : "CLUSTER" if mapping.clustered else "HOST",
                           })
                    for lc in lun_containers
                    for mapping in lc.luns
                    if mapping.volume_id == volume.id]
        else:
            return [Munch({self.ENTITY_TYPE.get_entity_name() : lc.name,
                           "lun" : mapping.lun,
                           "volume" : Volume.get_by_id(_api, mapping.volume_id).name,
                           "target_type" : "CLUSTER" if mapping.clustered else "HOST",
                           })
                    for lc in lun_containers
                    for mapping in lc.luns]

QUERY_FIELDS = strlist("name cluster luns fc_ports created_at")
SORTABLE_FIELDS = strlist("name created_at")

HOST_PORT_QUERY_FIELDS = strlist("host type port logged_in")

MAP_QUERY_FIELDS = strlist("host volume lun target_type")

@commandgroup(prefix="host", provides=(Host,))
class HostCommands(HostCommandsBase):
    ENTITY_TYPE = Host
    def _host_query_hook(self, hosts, cluster_names_by_id):
        return [Munch(name=host.name,
                      cluster=cluster_names_by_id[host.host_cluster_id] if host.host_cluster_id else None,
                      luns=len(host.luns),
                      fc_ports=fc_port_count,
                      created_at=host.created_at,
                      )
                for host in hosts
                for fc_port_count, iscsi_port_count in self._count_ports(host)]

    @command(
        description="Create a new host",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name):
        return self._create(_api, name=name)

    @command(
        description="Add port to host",
        types=(Host, WWN.List),
        substitute=dict(
            host="name",
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add_port(self, _api, host, ports):
        type = SCSIPortType.TYPES.FC
        with _api.aggregation() as aggregator:
            for port in ports:
                aggregator.next_task(port, "Added")
                self._post(_api, host, "ports", data=dict(
                    address=port,
                    type=type,
                    ))
        if len(ports) == 1:
            return '{} port "{}" added to host "{}"'.format(type.upper(), port, host.name)
        return '{} {} ports added to host "{}"'.format(len(ports), type.upper(), host.name)

    @command(
        description="Remove port from host",
        types=(Host, WWN.List),
        substitute=dict(
            host="name",
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove_port(self, _api, host, ports):
        type = SCSIPortType.TYPES.FC
        with _api.aggregation() as aggregator:
            for port in ports:
                aggregator.next_task(port, "Removed")
                self._del(_api, host, "ports", type, port)
        if len(ports) == 1:
            return '{} port "{}" removed from host "{}"'.format(type.upper(), port, host.name)
        return '{} {} ports removed from host "{}"'.format(len(ports), type.upper(), port, host.name)

    @command(
        description="Rename host",
        types=(Host, Name),
        substitute=dict(host="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def rename(self, _api, host, new_name):
        return self._rename(_api, host, new_name)

    @command(
        description="Delete host",
        types=(Host,),
        substitute=dict(host="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, host):
        return self._delete(_api, host)

    @command(
        description="Map {} to host".format(Volume.DESCRIPTION_SUBJECT),
        types=(Host, MasterVolume.List, VolumeSnap.List, VolumeClone.List, Lun),
        substitute=dict(host="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def map(self, _api, host, vol=None, snap=None, clone=None, lun=None):
        return self._map(_api, host, lun, vol, snap, clone)

    @command(
        description="Unmap {} from host".format(Volume.DESCRIPTION_SUBJECT),
        types=(Host, MasterVolume.List, VolumeSnap.List, VolumeClone.List, Lun),
        substitute=dict(host="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def unmap(self, _api, host, vol=None, snap=None, clone=None, lun=None):
        return self._unmap(_api, host, lun, vol, snap, clone)
    def _count_ports(self, host):
        fc, iscsi = 0, 0
        for port in host.ports:
            if port.type.lower() == "fc":
                fc += 1
            elif port.type.lower() == "iscsi":
                iscsi += 1
        return ((fc, iscsi),)

    @command(
        description="List existing hosts",
        types=(Host.List,),
        substitute=dict(host="name"),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS),
        )
    @render_collection(
        entity=Host,
        fields=QUERY_FIELDS,
        )
    def query(self, _api, host=None):
        return self._query(_api, host, _hook=partial(self._host_query_hook, cluster_names_by_id=LazyIdNameMapping(Cluster, _api)))

    @command(
        description="List ports of existing hosts",
        types=(Host, WWN),
        substitute=dict(host="name"),
        options=QueryOptions(fields=HOST_PORT_QUERY_FIELDS, detailed=True),
        )
    @render_collection(
        entity=Host,
        fields=HOST_PORT_QUERY_FIELDS,
        substitute=dict(host="name"),
        detailed_fields=HOST_PORT_QUERY_FIELDS,
        detailed_substitute=dict(host="name"),
        detailed_individually=True,
        )
    def port_query(self, _api, host=None, port=None):
        name = host.name if host else None
        hosts = self._query(_api, name=name)
        if port:
            port = port.replace(":", "")
            return [Munch(host=_host.name,
                          port=hostport.address,
                          type=hostport.type)
                    for _host in hosts
                    for hostport in _host.ports if port == hostport.address.replace(":", "")]
        else:
            return [Munch(host=_host.name,
                          port=hostport.address,
                          type=hostport.type)
                    for _host in hosts
                    for hostport in _host.ports]

    @command(
        description="List existing host mappings",
        types=(Host,),
        substitute=dict(host="name"),
        options=QueryOptions(fields=MAP_QUERY_FIELDS),
        )
    @render_collection(
        fields=MAP_QUERY_FIELDS,
        substitute=dict(
            target_type="mapping_type",
            ),
        )
    def map_query(self, _api, host=None):
        return self._map_query(_api, host)
