from __future__ import print_function
import __builtin__

### BEGIN PATCH FOR IMPORT FUNC ###
__original_import_func__ = __import__
import infinishell as __application_top_package_module__
def __custom_import__(name, *args, **kwargs):
    if not getattr(__application_top_package_module__, 'allow_direct_imports', False) \
    and name.startswith("infinishell.infinibox."):
        raise ImportError("""\
Cannot import infinishell.infinibox prior to application initialization.

If you're trying to import modules as part of a script, please add the following
lines to the beginning of your script:

import infinishell
infinishell.allow_direct_imports = True
""")
    return __original_import_func__(name, *args, **kwargs)
__builtin__.__import__ = __custom_import__
del __custom_import__
### END PATCH FOR IMPORT FUNC ###

# Import patches first
import infi.shell.utils.colorama_patches  # noqa

import sys
import traceback
import time
import socket
import itertools
import textwrap
import inspect
from optparse import OptionParser, IndentedHelpFormatter
from collections import namedtuple
from getpass import getpass
from requests import get
from requests.exceptions import SSLError, ConnectionError
from requests.packages.urllib3 import disable_warnings

from infi.shell.parser import CHARS as PARSER_CHARS
from infi.shell import (
    config, default_config, write_default_config, write_current_config, initialize_config,
    datadir,
    )

# Explicitly import config module to apply InfiniBox configuration
from . import infinishell_config  # noqa

from . import debug
from .version import VERSION
from .constants import (
    COMPANY_NAME_LOWER,
    PRODUCT_NAME,
    CLIENT_NAME,
    CLIENT_NAME_LOWER,
    CLIENT_NAME_UPPER,
    CLIENT_LONG_NAME,
    CLIENT_FULL_NAME,
    )
from .logging import configure_logging
from .environment import variables as env
from .utils import NamedDefaultDict, Version
from .encryption import encrypt_password, decrypt_password, is_encrypted_password
# from .plugins import plugins
from .exceptions import InvalidOptions

EXIT_MSG = "Goodbye!"
ENCODING = "utf-8"

DATA_DIR = "~/.{}/{}".format(COMPANY_NAME_LOWER, CLIENT_NAME_LOWER)
CONFIG_FILE = CLIENT_NAME_LOWER + "rc"

BASE_URI = "api"
REST_BASE_URI = "rest"
INTERNAL_BASE_URI = "internal"

COLOR_OPTIONS = ["always", "never", "auto"]
NO_VALUE = "__no_value__"
VERSION_PROBE_TIMEOUT_SECONDS = 10

FlagOption = namedtuple("FlagOption", ("name", "enabler_option", "disabler_option"))

env.set_prefix(CLIENT_NAME_UPPER)

def output(msg, *args, **kwargs):
    print(msg.format(*args, **kwargs), file=sys.stdout)

def output_stderr(msg, *args, **kwargs):
    print(msg.format(*args, **kwargs), file=sys.stderr)

def warn(*args, **kwargs):
    kwargs["__msg_prefix__"] = "WARNING: "
    error(*args, **kwargs)

def error(msg, *args, **kwargs):
    prefix = kwargs.pop("__msg_prefix__", "ERROR: ")
    prefixes = itertools.chain([prefix], itertools.cycle([" "*len(prefix)]))
    for line in textwrap.wrap(msg.format(*args, **kwargs), 70-len(prefix)):
        print(next(prefixes) + line, file=sys.stderr)

class InfiniShellOptionParser(OptionParser):
    def __init__(self, *args, **kwargs):
        OptionParser.__init__(self, *args, **kwargs)
        self._flag_options = []
    def error(self, msg):
        self.print_usage(sys.stderr)
        raise InvalidOptions(msg.capitalize())
    def add_flag_option(self, flag, enabler_options, enabler_help, disabler_options, disabler_help):
        default_value = getattr(default_config, flag)
        enabler_dest, disabler_dest = self._get_dest_name(enabler_options), self._get_dest_name(disabler_options)
        for options, help, dest, xor in ((enabler_options, enabler_help, enabler_dest, False),
                                         (disabler_options, disabler_help, disabler_dest, True)):
            is_enabled = default_value^xor
            self.add_option(*options, dest=dest, action="store_true", default=False,
                            help="{} (this is the default)".format(help) if is_enabled else help)
        self._flag_options.append(FlagOption(flag, enabler_dest, disabler_dest))
    def add_config_option(self, *args, **kwargs):
        assert args, "Option names are mandatory"
        assert "default" not in kwargs, "Defaults are not allowed in config options"
        key = kwargs.get("dest")
        key = key or kwargs.setdefault("dest", args[-1][2:] if args[-1].startswith("--") else args[-1][1:])
        if "help" in kwargs:
            kwargs["help"] = kwargs["help"].replace("%default", "{}").format(default_config[key])
        kwargs.setdefault("type", "string")
        if kwargs["type"] == "string":
            kwargs["default"] = NO_VALUE
        kwargs.update(
                action="callback",
                callback=self._config_option_callback,
                )
        self.add_option(*args, **kwargs)
    def _config_option_callback(self, option, option_str, value, parser):
        assert option.dest
        config[option.dest] = value
        setattr(parser.values, option.dest, value)
    def _get_dest_name(self, options):
        returned = None
        for option in options:
            if option[0:2] == "--":
                returned = option[2:]
                break
        return (returned or options[0][1:]).replace("-", "_")
    def set_options_in_config(self, values):
        for flag in self._flag_options:
            is_enabled = config[flag.name]
            if (is_enabled and getattr(values, flag.disabler_option) or
                not is_enabled and getattr(values, flag.enabler_option)):
                config[flag.name] = not is_enabled

def display_version_and_exit(*_, **__):
    output(CLIENT_FULL_NAME)
    sys.exit(0)

def write_default_config_and_exit(*_, **__):
    if datadir.exists(CONFIG_FILE):
        error("Configuration file already exists. Delete it first.")
    else:
        try:
            config_file_path = datadir.get_path(CONFIG_FILE)
            write_default_config(config_file_path)
            output('Default configuration file generated at "{}"'.format(config_file_path))
            sys.exit(0)
        except OSError as e:
            error("Failed to write configuration file: {}", e)
    sys.exit(2)

def list_env_vars_and_exit(*_, **__):
    def var_to_str(v):
        returned = "{}_{}".format(env.get_prefix(), v.name)
        if v.is_flag:
            returned += " (flag)"
        if v.help:
            returned += ": " + v.help
        if v.default is not None:
            returned += " (default: {})".format(v.default)
        return returned
    env_vars = "\n  ".join(var_to_str(v) for v in env.get_all() if not v.is_internal)
    output('Environment variables:\n  ' + env_vars)
    output('\nPossible flag values: ' + ', '.join(itertools.chain(env.Variable.POSITIVES,
                                                                env.Variable.NEGATIVES)))
    sys.exit(0)

def assert_options(options, args):
    if len(args) > 1:
        raise InvalidOptions("Too many arguments: {}".format(", ".join(repr(arg) for arg in args)), print_usage=True)
    if (options.command, options.file, options.command_help).count(None) < 2:
        raise InvalidOptions("Please specify at most one of the following arguments: -c, -f, -H")
    if options.color not in ([NO_VALUE] + COLOR_OPTIONS):
        raise InvalidOptions("Options for --color are: {}".format(", ".join(COLOR_OPTIONS)))

def handle_debug(options):
    if options.debug:
        debug.on()

def get_script(options):
    if options.command:
        script = options.command
        if PARSER_CHARS.cmd_separator in script:
            raise InvalidOptions("Multiple commands are only allowed in scripts")
    elif isinstance(options.command, basestring):
        raise InvalidOptions("No command was specified")
    elif options.file:
        with open(options.file) as f:
            script = f.read()
    elif options.command_help:
        script = "help on='{}'".format(options.command_help)
    elif not sys.stdin.isatty():
        script = sys.stdin.read()
    else:
        return None
    return script.decode(ENCODING)

def get_address(args):
    address = args[0] if args else env.get("HOSTNAME", config.default_address)
    if not address:
        address = raw_input("{} address: ".format(PRODUCT_NAME))
    address_port = address.split(":")
    address = address_port[0]
    if len(address_port) == 1:
        port = None
    elif len(address_port) == 2:
        port = int(address_port[1])
    else:
        raise InvalidOptions("Invalid address: {}".format(address))
    return address, port

def simplify_address(address):
    address = address.lower()
    simplified_address = "".join(c for c in address if c in ("abcdefghijklmnopqrstuvwxyz1234567890-_"))
    return "{}_{:x}".format(simplified_address, abs(hash(address)))

def get_history_filename(address):
    if not config.history:
        return None
    if config.history_file:
        return config.history_file
    simplified_address = simplify_address(address)
    return datadir.get_path("history", simplified_address)

def new_client_id():
    localhost = socket.gethostname()
    return "{}-{}-{:x}".format(CLIENT_NAME, localhost, int(time.time()*1000))

_login_password = None
def get_login_password():
    """ Invoke this function only once, to get the login password. """
    return globals().pop('_login_password')

def set_username_password(options):
    global _login_password
    if env.is_flagged("NO_AUTH"):
        return
    if options.username is NO_VALUE:
        if options.password is not None:
            raise InvalidOptions("Please specify either a username, a username and password, or neither.")
        # Username and password were not specified in options, so use config
        if config.password is not None:
            _login_password = decrypt_password(config.password)
    elif options.password is not None:
        _login_password = options.password
    # Remove potentially-clear-text password from memory
    del options.password
    def set_password():
        global _login_password
        _login_password = getpass("{} password: ".format(PRODUCT_NAME))
    if not config.username:
        auth = env.get("AUTH")
        if auth:
            auth = auth.split(":", 1)
            if len(auth) != 2:
                raise InvalidOptions('{}_AUTH environment variable should contain '
                                     'colon-separated username and password, e.g: "user:pass"'.format(CLIENT_NAME_UPPER))
            config.username, _login_password = auth
        else:
            if not sys.stdin.isatty():
                raise InvalidOptions("Credentials not set, and cannot get them while running "
                                     "non-interactively. Use the configuration file (see help), "
                                     "or {}_AUTH environment variable.".format(CLIENT_NAME_UPPER))
            config.username = raw_input("{} username: ".format(PRODUCT_NAME))
            set_password()
    elif not _login_password:
        if not sys.stdin.isatty():
            raise InvalidOptions("Only username was set in configuration file, and "
                                 "cannot get password while running non-interactively.")
        set_password()

def encrypt_password_in_config(config_file_path):
    if config.password is not None and not isinstance(config.password, basestring):
        error('Password in "{}" can be either a string, or null - ignoring', config_file_path)
        config.password = None
        return
    if config.password and not is_encrypted_password(config.password):
        config.password = encrypt_password(config.password)
        try:
            write_current_config(config_file_path)
        except Exception as e:
            error('Failed to encrypt password in configuration file: {}. PASSWORD REMAINS UNENCRYPTED!'.format(e))

def get_welcome_message(system_version):
    msg = "{}{} {}".format(
        PRODUCT_NAME, (" v" + system_version) if system_version else "", CLIENT_LONG_NAME)
    return "\n{}\n".format(msg)

def run(address, port, script, options):
    system_version = initialize_codebase_for_system(address, port)
    from infinishell.infinibox.main import run as run_func
    if env.is_flagged("DEBUG"):
        output_stderr("Code base: {}",
            # Intentionally parse the package name so we're sure it's right
            inspect.getmodule(run_func).__package__
                .replace("infinishell.infinibox", "")[1:]
                .replace("_", ".")
            or "latest")
    return run_func(NamedDefaultDict(
        address=address,
        port=port,
        script=script,
        encoding=ENCODING,
        welcome_msg=get_welcome_message(system_version),
        exit_msg=EXIT_MSG,
        **kwargify_options(options)
        ))

def initialize_codebase_for_system(address, port):
    """ Get system version, initialize codebase, and return system version. """
    system_version = get_system_version(address, port)  # May be None
    initialize_codebase(system_version)
    return system_version

def initialize_codebase(system_version):
    version_codebase = get_version_codebase(system_version)
    patch_import_func(version_codebase)

def get_system_version(address, port):
    assumed_version = (env.get("EMULATE_SYSTEM_VERSION") or "").strip()
    if assumed_version:
        return assumed_version
    if debug.is_on() or env.is_flagged("DEBUG"):
        output_stderr("Probing {}...", address)
    try:
        disable_warnings()
        response =  get(get_api_base_url(address, port) + "/rest/system/ping", timeout=VERSION_PROBE_TIMEOUT_SECONDS)
    except SSLError:
        response =  get(get_api_base_url(address, port) + "/rest/system/ping", timeout=VERSION_PROBE_TIMEOUT_SECONDS, verify=False)
    except ConnectionError:
        output_stderr("{} version could not be determined. Some {} commands may not be applicable.".format(PRODUCT_NAME, CLIENT_NAME))
        return None
    if not response.ok:
        output_stderr("{} version could not be determined. Some {} commands may not be applicable.".format(PRODUCT_NAME, CLIENT_NAME))
        return None
    version = response.headers.get("x-infinidat-version")
    return version and version.strip()

LATEST_CODEBASE = "infinibox"
VERSION_CODEBASE_MAPPING = (
    ("2.2.1", LATEST_CODEBASE),
    ("2.2.0", LATEST_CODEBASE),
    ("2.0.0", "infinibox_2_0"),
    )

def get_version_codebase(system_version):
    if not system_version:
        return LATEST_CODEBASE
    target_version = Version(system_version)
    for shell_version, codebase in VERSION_CODEBASE_MAPPING:
        if target_version >= Version(shell_version):
            return codebase
    warn("{} v{} does not support {} v{}".format(CLIENT_NAME, VERSION, PRODUCT_NAME, system_version))
    return LATEST_CODEBASE

def patch_import_func(version_codebase):
    """ Patch import statement according to the chosen version codebase. This
    results in """
    assert "infinishell.infinibox" not in sys.modules, "Codebase package is already in sys.modules"
    default_codebase_package = "infinishell.infinibox"
    this_package = __name__[:__name__.rfind('.')]
    version_codebase_package = "{}.{}".format(this_package, version_codebase)
    if version_codebase_package == default_codebase_package:
        __builtin__.__import__ = __original_import_func__
    def custom_import(name, globals=None, locals=None, fromlist=(), level=-1):
        # Commented out code that handles relative imports, since it's a bit
        # broken... Anyway there's no use case for relative imports right now.
        # if level >= 0:
        #     assert globals
        #     package = globals.get("__package__") or globals.get("__name__")
        #     assert package
        #     name = importlib._resolve_name(name, package, level)
        #     level = -1
        if name.startswith(default_codebase_package):
            if name == default_codebase_package:
                name = version_codebase_package
            elif name[len(default_codebase_package)] == ".":
                name = version_codebase_package + name[len(default_codebase_package):]
        return __original_import_func__(name, globals, locals, fromlist, level)
    __builtin__.__import__ = custom_import

def get_api_base_url(address, port):
    ssl_enabled = config.api_ssl and not env.is_flagged("NO_SSL")
    return "{protocol}://{address}:{port}/{base_uri}".format(
        protocol="https" if ssl_enabled else "http",
        address=address,
        port=port or (config.api_ssl_port if ssl_enabled else config.api_port),
        base_uri=BASE_URI,
        )

def kwargify_options(options):
    return {attr : getattr(options, attr)
            for attr in (a for a in dir(options) if a[0] != "_")}

def _main():
    datadir.init(DATA_DIR)
    config_file_path = datadir.get_path(CONFIG_FILE)
    initialize_config(config_file_path)
    op = InfiniShellOptionParser(prog=CLIENT_NAME_LOWER, usage="Usage: %prog [options] system[:port]",
                                 epilog='Please note that defaults may be overridden in the configuration '
                                        'file: "{}". Also, some environment variables are available for '
                                        'your convenience. To list them, use the --list-environment option.'
                                        .format(config_file_path),
                                 formatter=IndentedHelpFormatter(max_help_position=30))
    op.add_option("-v", "--version", action="callback", callback=display_version_and_exit,
                  help="Print version and exit")
    op.add_option("-c", "--cmd", dest="command", metavar="CMD", default=None,
                  help="Commands to execute (separated by semicolons)")
    op.add_option("-f", "--file", metavar="FILE", default=None,
                  help="Script file to execute")
    op.add_option("-d", "--debug", action="store_true", default=False,
                  help="Enable debug prints")
    op.add_option("-r", "--raw", action="store_true", default=False,
                  help="Display raw API responses instead of normal rendering")
    op.add_config_option("-u", "--user", dest="username",
                         help="Login username")
    op.add_option("-p", "--password",
                  help="Login password")
    op.add_option("-H", dest="command_help", metavar="CMD", default=None,
                  help="Display syntax of the specified command")
    op.add_config_option("-l", "--log", dest="log_file", metavar="FILE",
                         help="Log commands to the specified file (default: %default)")
    op.add_option("--write-default-config", action="callback", callback=write_default_config_and_exit,
                  help='Generate the default configuration file at "{}"'.format(config_file_path))
    op.add_option("--list-environment", action="callback", callback=list_env_vars_and_exit,
                  help="List environment variables")
    op.add_config_option("--color", metavar="WHEN",
                         help="Whether or not to output with color. Options are: always, never, auto (default: %default)")
    op.add_flag_option(flag="keep_going",
                       enabler_options=("-k", "--keep-going"),
                       enabler_help="Keep going when administration errors occur (syntax errors always abort the entire script)",
                       disabler_options=("--fail-fast",),
                       disabler_help="Stop running script when the first administration error occurs")
    op.add_flag_option(flag="quiet",
                       enabler_options=("-q", "--quiet"),
                       enabler_help="No output apart from errors",
                       disabler_options=("--not-quiet",),
                       disabler_help="Normal output, the opposite of --quiet")
    op.add_flag_option(flag="paging",
                       enabler_options=("--with-paging",),
                       enabler_help="Enable paging of large outputs",
                       disabler_options=("--no-paging",),
                       disabler_help="Disable paging of large outputs")
    op.add_flag_option(flag="auto_confirm_prompts",
                       enabler_options=("-y", "--yes"),
                       enabler_help="Automatically confirm prompts (of potentially destructive commands)",
                       disabler_options=("--no",),
                       disabler_help="Automatically reject prompts (of potentially destructive commands)")
    op.add_config_option("--history-file", metavar="PATH", dest="history_file",
                         help="Path of history file, for cross-session history of commands. By default this "
                              "is saved in {}'s data directory, under a unique name generated off of "
                              "the target system address".format(CLIENT_NAME))
    op.add_flag_option(flag="history",
                       enabler_options=("--with-history",),
                       enabler_help="Enable cross-session history of commands",
                       disabler_options=("--no-history",),
                       disabler_help="Disable cross-session history of commands")
    op.add_flag_option(flag="csv_output",
                       enabler_options=("--csv",),
                       enabler_help="CSV output mode (optimize output for easy parsing, see RFC 4180)",
                       disabler_options=("--no-csv",),
                       disabler_help="Disable CSV output mode")
    op.add_flag_option(flag="json_output",
                       enabler_options=("--json",),
                       enabler_help="JSON output mode (optimize output for easy parsing, see RFC 4180)",
                       disabler_options=("--no-json",),
                       disabler_help="Disable JSON output mode")
    op.add_config_option("--request-timeout", dest="api_request_timeout_seconds", metavar="SECONDS",
                         type="int",
                         help="Maximal amount of time to wait for management response (default: %default)")
    op.add_flag_option(flag="api_ssl",
                       enabler_options=("--with-ssl",),
                       enabler_help="Encrypt connection using SSL or TLS",
                       disabler_options=("--no-ssl",),
                       disabler_help="Do not encrypt connection")
    try:
        encrypt_password_in_config(config_file_path)  # Important: Do this before parsing args, as it writes the config file!
        options, args = op.parse_args()
        assert_options(options, args)
        handle_debug(options)
        op.set_options_in_config(options)
        address, port = get_address(args)
        script = get_script(options)
        set_username_password(options)
        configure_logging(encoding=ENCODING)
        return run(address, port, script, options)
    except KeyboardInterrupt:
        error("Aborted")
        return 1
    except Exception as e:
        if isinstance(e, InvalidOptions):
            if e.print_help:
                op.print_help()
            elif e.print_usage:
                op.print_usage()
        if env.is_flagged("DEBUG"):
            traceback.print_exc()
        else:
            error(str(e))
        return 1

def main():
    if env.is_flagged("PROFILER"):
        import cProfile
        cProfile.run("from infinishell.main import _main; _main()", sort="cumulative")
        return 0
    else:
        return _main()

if __name__ == '__main__':
    sys.exit(main())
